# Forlix SrcDS Query v2.0 - 2010/05/22
# http://forlix.org/, df@forlix.org
#
# Copyright (c) 2008-2010 Dominik Friedrichs

use strict;
use Socket;
use Encode qw(decode);

#use warnings;
#use CGI::Carp qw(fatalsToBrowser);

use constant
{
  'TIMEOUT',      0.2,
  'RETRY',        3,

  'PKT_HEAD',                 "\xFF\xFF\xFF\xFF",
  'PKT_FRGHEAD',              "\xFE\xFF\xFF\xFF",

  'A2S_INFO_QUERY',           "\x54Source Engine Query\x00",
  'A2S_PLAYER_QUERY',         "\x55",

  'A2S_INFO_RESPONSE',        "\x49",
  'A2S_PLAYER_RESPONSE',      "\x44",
  'A2S_CHALLENGE_RESPONSE',   "\x41",
};

sub utf8_html(\$);
sub recv_tm(*\$*);

sub a2s_query(\$\$\%\%);

sub parse_A2S_INFO_RESPONSE(\$\%);
sub parse_A2S_PLAYER_RESPONSE(\$\%);


sub utf8_html(\$)
{
  my ($str) = @_;
  $$str = decode('utf8', $$str, 0);
  $$str =~ tr/\x{FFFD}/\x{2026}/;
  $$str =~ s/([^a-zA-Z0-9\-\+\.\[\]\(\)\{\}\|\_\:\!\*\?\s])/'&#'.unpack('U0U*',$1).';'/eg;
  $$str =~ s/^\s/&#160;/;
  $$str =~ s/\s\s/&#160;&#160;/g;
  return(1);
}


sub recv_tm(*\$*)
{
  my ($sock, $pkt, $tmout) = @_;
  my $rin = 0;
  vec($rin, fileno($sock), 1) = 1;

  my ($nf, $tleft) = select($rin, undef, undef, $tmout);
  return(0)
    if(!$nf);

  recv($sock, $$pkt, 2048, 0);
  return($tleft);
}


sub a2s_query(\$\$\%\%)
{
  my ($ip, $port, $a2sinfo, $a2splayer) = @_;
  my $sockaddr = sockaddr_in($$port, inet_aton($$ip));

  socket(SOCK_QINFO, PF_INET, SOCK_DGRAM, getprotobyname('udp'));
  socket(SOCK_QPLYR, PF_INET, SOCK_DGRAM, getprotobyname('udp'));

  my $query = 0;
  my $challenge = 0;

  query:
  {
    $challenge = defined($$a2splayer{'challenge'})? $$a2splayer{'challenge'} : 0;

    %$a2sinfo = ();
    %$a2splayer = ();

    my $pkt_a2sinfo = 0;
    my $pkt_a2splayer = 0;

    for(my $r = 0;
        $r <= RETRY && (!$pkt_a2sinfo || !$pkt_a2splayer);
        $r++)
    {
      my $timeleft = TIMEOUT;

      send(SOCK_QINFO, (PKT_HEAD.A2S_INFO_QUERY), 0, $sockaddr)
        if(!$pkt_a2sinfo);
      send(SOCK_QPLYR, (PKT_HEAD.A2S_PLAYER_QUERY.pack('N', $challenge)), 0, $sockaddr)
        if(!$pkt_a2splayer);

      $timeleft = recv_tm(SOCK_QINFO, $pkt_a2sinfo, $timeleft)
        if(!$pkt_a2sinfo);
      $timeleft = recv_tm(SOCK_QPLYR, $pkt_a2splayer, $timeleft)
        if(!$pkt_a2splayer);
    }

    if(!$pkt_a2sinfo || !$pkt_a2splayer)
    {
      $$a2sinfo{'error'} = 'Server timed out';
      return(0);
    }

    my $a2sinfo_return = parse_A2S_INFO_RESPONSE($pkt_a2sinfo, %$a2sinfo);
    my $a2splayer_return = parse_A2S_PLAYER_RESPONSE($pkt_a2splayer, %$a2splayer);

    if(!$a2sinfo_return || !$a2splayer_return)
    {
      %$a2sinfo = ();
      %$a2splayer = ();

      $$a2sinfo{'error'} = sprintf('Failed to parse server response (%i; %i)',
                                   $a2sinfo_return, $a2splayer_return);
      return(0);
    }

    if($a2splayer_return < 0)
    # got a new challenge instead of player response
    {
      if(++$query > 1)
      {
        %$a2sinfo = ();
        %$a2splayer = ();

        $$a2sinfo{'error'} = sprintf('Fresh challenge invalid (%X)', $challenge);
        return(0);
      }

      goto query;
    }
  }

  close(SOCK_QINFO);
  close(SOCK_QPLYR);

  $$a2splayer{'challenge'} = $challenge;
  return(1);
}


sub parse_A2S_INFO_RESPONSE(\$\%)
{
  my ($pkt, $prs) = @_;

  return(0)
    if(substr($$pkt, 0, 4) ne PKT_HEAD
    || substr($$pkt, 4, 1) ne A2S_INFO_RESPONSE);

  my $off = 5;
  my $tmp = 0;

  $$prs{'netver'} = unpack('C', substr($$pkt, $off++, 1));

  $tmp = $off;
  $off = index($$pkt, "\x00");
  $$prs{'name'} = substr($$pkt, $tmp, $off++ - $tmp);
  utf8_html($$prs{'name'});

  $tmp = $off;
  $off = index($$pkt, "\x00", $off);
  $$prs{'map'} = substr($$pkt, $tmp, $off++ - $tmp);

  $tmp = $off;
  $off = index($$pkt, "\x00", $off);
  $$prs{'game'} = substr($$pkt, $tmp, $off++ - $tmp);

  $tmp = $off;
  $off = index($$pkt, "\x00", $off);
  $$prs{'gamedesc'} = substr($$pkt, $tmp, $off++ - $tmp);
  $$prs{'appid'} = unpack('S', substr($$pkt, $off, 2));
  $off += 2;

  $$prs{'numplayers'} = unpack('C', substr($$pkt, $off++, 1));
  $$prs{'maxplayers'} = unpack('C', substr($$pkt, $off++, 1));
  $$prs{'bots'} = unpack('C', substr($$pkt, $off++, 1));
  $$prs{'dedicated'} = substr($$pkt, $off++, 1);
  $$prs{'system'} = substr($$pkt, $off++, 1);
  $$prs{'password'} = unpack('C', substr($$pkt, $off++, 1));
  $$prs{'secure'} = unpack('C', substr($$pkt, $off++, 1));

  $tmp = $off;
  $off = index($$pkt, "\x00", $off);
  $$prs{'gamever'} = substr($$pkt, $tmp, $off++ - $tmp);

  return(1);
}


sub parse_A2S_PLAYER_RESPONSE(\$\%)
{
  my ($pkt, $prs) = @_;

  return(0)
    if(substr($$pkt, 0, 4) ne PKT_HEAD);

  my $response = substr($$pkt, 4, 1);

  if($response eq A2S_CHALLENGE_RESPONSE)
  {
    $$prs{'challenge'} = unpack('N', substr($$pkt, 5, 4));
    return(-1);
  }

  if($response eq A2S_PLAYER_RESPONSE)
  {
    my $off = 5;
    my $tmp = 0;

    my $pkt_len = length($$pkt);
    my $numplayers = unpack('C', substr($$pkt, $off++, 1));

    $$prs{'sets'} = 0;
    my @tmp;

    while($off < $pkt_len)
    {
      my $index = unpack('C', substr($$pkt, $off++, 1));
      $tmp = $off;
      $off = index($$pkt, "\x00", $off);

      my $player = substr($$pkt, $tmp, $off++ - $tmp);
      utf8_html($player);

      my $score = unpack('l', substr($$pkt, $off, 4));
      $off += 4;

      my $time = unpack('f', substr($$pkt, $off, 4));
      $time = sprintf('%.f', $time);
      $off += 4;

      next
        if(!$player);

      push(@tmp, [$player, $score, $time]);
      $$prs{'sets'}++;
    }

    my $index = 0;
    for my $player (sort{$b->[2] <=> $a->[2]} @tmp)
    {
      $$prs{$index++} = @{$player}[0].','.@{$player}[1].','.@{$player}[2];
    }

    return(1);
  }

  return(0);
}

1;
